/*  XRoar - a Dragon/Tandy Coco emulator
 *  Copyright (C) 2003-2004  Ciaran Anscomb
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <SDL/SDL.h>

#include "types.h"
#include "keyboard.h"
#include "pia.h"
#include "snapshot.h"
#include "xroar.h"
#include "logging.h"
#include "ui.h"
#include "wd2797.h"

#define KEY_PRESSED(s) keyboard_column[keymap[s][0]] &= ~(1<<keymap[s][1]); \
	keyboard_row[keymap[s][1]] &= ~(1<<keymap[s][0])
#define KEY_RELEASED(s) keyboard_column[keymap[s][0]] |= 1<<keymap[s][1]; \
	keyboard_row[keymap[s][1]] |= 1<<keymap[s][0]

static Uint32 control = 0;

void keyboard_init_arch(void) {
}

void keyboard_poll(void) {
	SDL_Event event;
	SDLKey sym;
	while (SDL_PollEvent(&event) == 1) {
		switch(event.type) {
			case SDL_KEYDOWN:
				sym = event.key.keysym.sym;
				switch (sym) {
					case SDLK_UP: KEY_PRESSED(11); break;
					case SDLK_DOWN: KEY_PRESSED(10); break;
					case SDLK_LEFT: KEY_PRESSED(8); break;
					case SDLK_RIGHT: KEY_PRESSED(9); break;
					case SDLK_LCTRL: case SDLK_RCTRL:
							 control = 1; break;
					case SDLK_LSHIFT: case SDLK_RSHIFT:
							 KEY_PRESSED(0); break;
					default: if (sym == SDLK_c && control) exit(0);
						if (sym == SDLK_s && control) { snapshot_state = 1; break; }
						if (sym == SDLK_l && control) { snapshot_state = 2; break; }
						if (sym == SDLK_r && control) { snapshot_state = 3; break; }
						if (sym == SDLK_e && control) { dragondos_enabled = !dragondos_enabled; break; }
						if (sym == SDLK_d && control) { wd2797_load_disk(ui_get_filename("VDK"), 0); break; }
						if (sym == SDLK_m && control) { machine_set_romtype(machine_romtype+1); snapshot_state = 3; break; }
						if (sym == SDLK_k && control) { machine_set_keymap(machine_keymap+1); break; }
#ifdef TRACE
						//if (sym == SDLK_p && control) { xroar_dumpstate(); break; }
						if (sym == SDLK_t && control) { trace ^= 1; break; }
#endif
						if (sym < 128 && keymap[sym][0] != 8) {
							KEY_PRESSED(sym);
						}
				}
				break;
			case SDL_KEYUP:
				sym = event.key.keysym.sym;
				switch (sym) {
					case SDLK_UP: KEY_RELEASED(11); break;
					case SDLK_DOWN: KEY_RELEASED(10); break;
					case SDLK_LEFT: KEY_RELEASED(8); break;
					case SDLK_RIGHT: KEY_RELEASED(9); break;
					case SDLK_LCTRL: case SDLK_RCTRL:
							 control = 0; break;
					case SDLK_LSHIFT: case SDLK_RSHIFT:
							 KEY_RELEASED(0); break;
					default:
						if (sym < 128 && keymap[sym][0] != 8) {
							KEY_RELEASED(sym);
						}
						break;
				}
				break;
			default:
				break;
		}
	}
	keyboard_column_update();
	keyboard_row_update();
}
