/*  XRoar - a Dragon/Tandy Coco emulator
 *  Copyright (C) 2003-2004  Ciaran Anscomb
 *
 *  See COPYING for redistribution conditions. */

#ifndef __MACHINE_H__
#define __MACHINE_H__

#include "types.h"

/* Dragon 64s and later Dragon 32s used a 14.218MHz crystal
 * (ref: "Dragon 64 differences", Graham E. Kinns and then a motherboard) */
#define OSCILLATOR_RATE 14218000

/* NTSC timings: */
/*
#define CYCLES_PER_SCANLINE 910
#define ACTIVE_SCANLINES_PER_FRAME 242
#define TOTAL_SCANLINES_PER_FRAME 262
#define TOP_BORDER_OFFSET 25
*/

/* NTSC kludged timings: */
#define CYCLES_PER_SCANLINE 910
#define ACTIVE_SCANLINES_PER_FRAME 262
#define TOTAL_SCANLINES_PER_FRAME 310
#define TOP_BORDER_OFFSET 25

#define CPU_RATE_DIVISOR 16
#define CPU_RATE (OSCILLATOR_RATE / CPU_RATE_DIVISOR)

#define MACHINE_DRAGON 0
#define MACHINE_COCO 1

#define MACHINE_KEYMAP_DRAGON 0
#define MACHINE_KEYMAP_COCO 1

extern Uint32 machine_romtype, machine_keymap;
extern Uint32 dragondos_enabled;
extern Uint32 brk_csrdon, brk_bitin;
extern Uint32 keymap[128][2];
extern Uint8 ram0[0x8000];
extern Uint8 ram1[0x8000];
extern Uint8 rom0[0x8000];

void machine_init(void);
void machine_reset(void);
void machine_set_romtype(Uint32 mode);
void machine_set_keymap(Uint32 mode);

#endif  /* __MACHINE_H__ */
